/*
 * Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
 *
 * Licensed under the Apache License, Version 2.0 with LLVM exception (the
 * "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License and of the exception at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     https://spdx.org/licenses/LLVM-exception.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

use ::clap::Parser;

enum ExitCode {
    Ok(),
    Generic(),
    CompileFailed(),
}

impl Into<::std::process::ExitCode> for ExitCode {
    fn into(self) -> ::std::process::ExitCode {
        match self {
            ExitCode::Ok() => ::std::process::ExitCode::SUCCESS,
            ExitCode::Generic() => ::std::process::ExitCode::from(1),
            ExitCode::CompileFailed() => ::std::process::ExitCode::from(1),
        }
    }
}

/// Main arguments for bok
#[derive(::clap::Parser, Debug)]
#[command(name="bok", about="tools to build and configure systems",
    version, about, long_about = None)]
struct Cli {
    /// subcommand
    #[command(subcommand)]
    command: MainCmd,
    /// log level of bok messages
    #[arg(long, default_value = "info")]
    log: LogLevel,
}

/// Available subcommands for `bok`
#[derive(Debug, ::clap::Subcommand)]
enum MainCmd {
    /// Build
    Build {
        #[arg(long, short)]
        path: Option<::std::ffi::OsString>,
        #[arg(long, short, group = "target")]
        example: Option<String>,
        #[arg(long, short, group = "target")]
        lib: Option<String>,
    },
    /// Repository stuff
    Repo(RepoArgs),
}

#[derive(
    ::clap::ValueEnum, ::strum::Display, Copy, Clone, Debug, PartialEq, Eq,
)]
enum LogLevel {
    Error,
    Warn,
    Info,
    Debug,
    Trace,
}

impl Into<::slog::Level> for LogLevel {
    fn into(self) -> ::slog::Level {
        match self {
            LogLevel::Error => ::slog::Level::Error,
            LogLevel::Warn => ::slog::Level::Warning,
            LogLevel::Info => ::slog::Level::Info,
            LogLevel::Debug => ::slog::Level::Debug,
            LogLevel::Trace => ::slog::Level::Trace,
        }
    }
}

/// Handle a repository
#[derive(Debug, ::clap::Args)]
struct RepoArgs {
    #[arg(required = true)]
    name: ::bok::RepoName,
    #[command(subcommand)]
    commands: RepoCommands,
}

#[derive(Debug, ::clap::Subcommand)]
enum RepoCommands {
    List {},
}

fn main() -> ::std::process::ExitCode {
    let args = Cli::parse();

    let log = {
        use ::slog::Drain;
        let decorator = ::slog_term::TermDecorator::new().build();
        let drain = ::slog_term::FullFormat::new(decorator).build().fuse();
        let drain = ::slog_async::Async::new(drain).build().fuse();
        let drain = ::slog::LevelFilter::new(drain, args.log.into()).fuse();

        let log = ::slog::Logger::root(drain, ::slog::o!());
        log
    };

    match args.command {
        MainCmd::Build { path, example, lib } => {
            build(log, path, example, lib).into()
        }
        MainCmd::Repo(_) => {
            ::slog::info!(log, "got repo cmd");
            ExitCode::Generic().into()
        }
    }
}

fn build(
    log: ::slog::Logger,
    path: Option<::std::ffi::OsString>,
    example: Option<String>,
    lib: Option<String>,
) -> ExitCode {
    if let Some(dir) = path {
        if let Err(_) = ::std::env::set_current_dir(&dir) {
            ::slog::error!(log, "can't change dir to {:?}", dir);
            return ExitCode::Generic();
        };
        ::slog::debug!(log, "changed dir to {:?}", dir);
    };

    if !::std::path::Path::new("./Cargo.toml").exists() {
        ::slog::error!(
            log,
            "path does not have the 'Cargo.toml' file. Can't build"
        );
        return ExitCode::Generic();
    }

    let mut cargo = ::std::process::Command::new("cargo");
    cargo.arg("build");
    if let Some(example_name) = example {
        cargo.arg("--example").arg(example_name);
    }
    if let Some(lib_name) = lib {
        cargo.arg("--lib").arg(lib_name);
    }

    cargo.stderr(::std::process::Stdio::null());
    cargo.stdout(::std::process::Stdio::null());
    let res = cargo.status();
    match res {
        Ok(code) => {
            if code.success() {
                ::slog::info!(log, "compilation successful");
                ExitCode::Ok()
            } else {
                ::slog::error!(log, "cargo exited with error");
                ExitCode::CompileFailed()
            }
        }
        Err(_) => {
            ::slog::error!(log, "can't run cargo");
            ExitCode::CompileFailed()
        }
    }
}
